#![deny(missing_docs)]
#![warn(clippy::all)]

//! A crate for small integer optimization. Provides the [`SmallInt`] type. When possible this will
//! inline an integer and store it on the stack if that integer is small. However, for larger values,
//! this will be instead stored on the heap as a pointer to a `u32` slice, a length, and a sign.

// Invariant: If a small integer is within the bounds of an inline value, it must be inline.
// Invariant: If a small integer is on the heap, the size is the minimum digits required to
// represent it.

#[cfg(feature = "num-bigint")]
use num_bigint::BigInt;

#[cfg(feature = "num-bigint")]
use num_bigint::BigUint;

#[cfg(feature = "num-bigint")]
use num_bigint::Sign;

use core::mem::ManuallyDrop;

/// An error that occurred when processing a `SmallInt`.
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum SmallIntError {
    /// Conversion error when converting from SmallInt to other integer types.
    ConversionError,
}

impl std::error::Error for SmallIntError {}

impl core::fmt::Display for SmallIntError {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        let Self::ConversionError = self;
        write!(f, "Error converting integer to SmallInt.")
    }
}

/// An integer-like type that will store small integers up to `i128` inline. Larger integers are
/// represented as a slice to a sequence of base 2<sup>32</sup> digits represented as a `*mut u32`.
#[derive(Clone, PartialEq, Eq)]
pub struct SmallInt(SmallIntType);

/// An integer-like type that will store small integers up to `u128` inline. Larger integers are
/// represented as a slice to a sequence of base 2<sup>32</sup> digits represented as a `*mut u32`.
#[derive(Clone, PartialEq, Eq)]
pub struct SmallUint(SmallUintType);

#[derive(Clone, PartialEq, Eq)]
enum SmallIntType {
    Inline(i128),
    Heap((*mut u32, isize)),
}

#[derive(Clone, PartialEq, Eq)]
enum SmallUintType {
    Inline(u128),
    Heap((*mut u32, usize)),
}

impl Drop for SmallInt {
    fn drop(&mut self) {
        if let Self(SmallIntType::Heap((r, s))) = self {
            let size = usize::try_from(s.abs()).unwrap();
            let slice = unsafe { core::slice::from_raw_parts_mut(*r, size) };
            unsafe { std::mem::drop(Box::from_raw(slice)) }
        }
    }
}

impl Drop for SmallUint {
    fn drop(&mut self) {
        if let Self(SmallUintType::Heap((r, s))) = self {
            let slice = unsafe { core::slice::from_raw_parts_mut(*r, *s) };
            unsafe { std::mem::drop(Box::from_raw(slice)) }
        }
    }
}

macro_rules! int_impl {
    ($itype:ty, $rt:tt, $rtt:tt, $n:tt) => {
        impl From<$itype> for $rt {
            fn from(a: $itype) -> Self {
                Self(<$rtt>::Inline($n::from(a)))
            }
        }

        impl TryFrom<$rt> for $itype {
            type Error = SmallIntError;

            fn try_from(s: $rt) -> Result<Self, Self::Error> {
                match s.0 {
                    $rtt::Inline(i) => {
                        <$itype>::try_from(i).map_err(|_| SmallIntError::ConversionError)
                    }
                    $rtt::Heap((_, _)) => Err(SmallIntError::ConversionError),
                }
            }
        }
    };
}

int_impl!(u8, SmallInt, SmallIntType, i128);
int_impl!(u16, SmallInt, SmallIntType, i128);
int_impl!(u32, SmallInt, SmallIntType, i128);
int_impl!(u64, SmallInt, SmallIntType, i128);
int_impl!(i8, SmallInt, SmallIntType, i128);
int_impl!(i16, SmallInt, SmallIntType, i128);
int_impl!(i32, SmallInt, SmallIntType, i128);
int_impl!(i64, SmallInt, SmallIntType, i128);
int_impl!(i128, SmallInt, SmallIntType, i128);

int_impl!(u8, SmallUint, SmallUintType, u128);
int_impl!(u16, SmallUint, SmallUintType, u128);
int_impl!(u32, SmallUint, SmallUintType, u128);
int_impl!(u64, SmallUint, SmallUintType, u128);
int_impl!(u128, SmallUint, SmallUintType, u128);

macro_rules! try_from_itou {
    ($itype:ty) => {
        impl TryFrom<$itype> for SmallUint {
            type Error = SmallIntError;

            fn try_from(a: $itype) -> Result<Self, Self::Error> {
                Ok(Self(SmallUintType::Inline(
                    u128::try_from(a).map_err(|_| SmallIntError::ConversionError)?,
                )))
            }
        }

        impl TryFrom<SmallUint> for $itype {
            type Error = SmallIntError;

            fn try_from(s: SmallUint) -> Result<Self, Self::Error> {
                match s.0 {
                    SmallUintType::Inline(i) => {
                        <$itype>::try_from(i).map_err(|_| SmallIntError::ConversionError)
                    }
                    SmallUintType::Heap((_, _)) => Err(SmallIntError::ConversionError),
                }
            }
        }
    };
}

try_from_itou!(i8);
try_from_itou!(i16);
try_from_itou!(i32);
try_from_itou!(i64);
try_from_itou!(i128);

impl From<u128> for SmallInt {
    fn from(a: u128) -> Self {
        match i128::try_from(a) {
            Ok(i) => Self(SmallIntType::Inline(i)),
            Err(_) => {
                let mut v = a;
                let mut vec = Vec::with_capacity(4);
                while v != 0 {
                    vec.push(v as u32);

                    v >>= 32;
                }
                let mut slice = ManuallyDrop::new(vec.into_boxed_slice());
                Self(SmallIntType::Heap((
                    slice.as_mut_ptr(),
                    isize::try_from(slice.len()).unwrap(),
                )))
            }
        }
    }
}

impl TryFrom<SmallInt> for u128 {
    type Error = SmallIntError;

    fn try_from(s: SmallInt) -> Result<Self, Self::Error> {
        match s.0 {
            SmallIntType::Inline(i) => {
                u128::try_from(i).map_err(|_| SmallIntError::ConversionError)
            }
            SmallIntType::Heap((r, s)) => {
                let mut ret: u128 = 0;
                let mut bits = 0;
                let size = usize::try_from(s.abs()).unwrap();
                let slice = unsafe { core::slice::from_raw_parts(r, size) };
                for i in slice {
                    if bits >= 128 {
                        return Err(SmallIntError::ConversionError);
                    }
                    ret |= u128::from(*i) << bits;
                    bits += 32;
                }

                Ok(ret)
            }
        }
    }
}

impl From<SmallUint> for SmallInt {
    fn from(s: SmallUint) -> Self {
        match s.0 {
            SmallUintType::Inline(i) => SmallInt::from(i),
            SmallUintType::Heap((r, s)) => {
                SmallInt(SmallIntType::Heap((r, isize::try_from(s).unwrap())))
            }
        }
    }
}

impl TryFrom<SmallInt> for SmallUint {
    type Error = SmallIntError;

    fn try_from(value: SmallInt) -> Result<Self, Self::Error> {
        match value.0 {
            SmallIntType::Inline(i) => Self::try_from(i),
            SmallIntType::Heap((r, s)) => {
                let size = usize::try_from(s).map_err(|_| SmallIntError::ConversionError)?;
                if size > 4 {
                    Ok(Self(SmallUintType::Heap((r, size))))
                } else {
                    Ok(Self(SmallUintType::Inline(u128::try_from(value)?)))
                }
            }
        }
    }
}

#[cfg(feature = "num-bigint")]
impl From<BigInt> for SmallInt {
    fn from(b: BigInt) -> Self {
        match (&b).try_into() {
            Ok(i) => Self(SmallIntType::Inline(i)),
            Err(_) => {
                let (sign, vec) = b.to_u32_digits();
                let mut slice = ManuallyDrop::new(vec.into_boxed_slice());
                let size = match sign {
                    Sign::Minus => -isize::try_from(slice.len()).unwrap(),
                    Sign::NoSign => panic!(
                        "Shouldn't happen; BigInts which store zero should convert to inline."
                    ),
                    Sign::Plus => isize::try_from(slice.len()).unwrap(),
                };
                Self(SmallIntType::Heap((slice.as_mut_ptr(), size)))
            }
        }
    }
}

#[cfg(feature = "num-bigint")]
impl From<SmallInt> for BigInt {
    fn from(s: SmallInt) -> Self {
        match s.0 {
            SmallIntType::Inline(i) => Self::from(i),
            SmallIntType::Heap((r, s)) => {
                let size = usize::try_from(s.abs()).unwrap();
                let sign = s.signum();
                let slice = unsafe { core::slice::from_raw_parts(r, size) };
                let bs = match sign {
                    0 => Sign::NoSign,
                    -1 => Sign::Minus,
                    1 => Sign::Plus,
                    _ => panic!("This shouldn't happen; There are only 3 signums."),
                };

                BigInt::new(bs, slice.to_vec())
            }
        }
    }
}

#[cfg(feature = "num-bigint")]
impl From<BigUint> for SmallUint {
    fn from(b: BigUint) -> Self {
        match (&b).try_into() {
            Ok(i) => Self(SmallUintType::Inline(i)),
            Err(_) => {
                let vec = b.to_u32_digits();
                let mut slice = ManuallyDrop::new(vec.into_boxed_slice());
                let size = slice.len();
                Self(SmallUintType::Heap((slice.as_mut_ptr(), size)))
            }
        }
    }
}

#[cfg(feature = "num-bigint")]
impl From<SmallUint> for BigUint {
    fn from(s: SmallUint) -> Self {
        match s.0 {
            SmallUintType::Inline(i) => Self::from(i),
            SmallUintType::Heap((r, s)) => {
                let slice = unsafe { core::slice::from_raw_parts(r, s) };

                BigUint::new(slice.to_vec())
            }
        }
    }
}

#[cfg(test)]
mod conversion_tests {

    use crate::SmallInt;

    #[cfg(feature = "num-bigint")]
    use num_bigint::{BigInt, Sign};

    macro_rules! conversion_tests {
        ($t:ty, $i:ident) => {
            #[test]
            fn $i() {
                let i = <$t>::MAX;
                let s = SmallInt::from(i);
                assert_eq!(<$t>::try_from(s).unwrap(), i);

                let i = <$t>::MIN;
                let s = SmallInt::from(i);
                assert_eq!(<$t>::try_from(s).unwrap(), i);
            }
        };
    }

    conversion_tests!(u8, test_u8);
    conversion_tests!(i8, test_i8);
    conversion_tests!(u16, test_u16);
    conversion_tests!(i16, test_i16);
    conversion_tests!(u32, test_u32);
    conversion_tests!(i32, test_i32);
    conversion_tests!(u64, test_u64);
    conversion_tests!(i64, test_i64);
    conversion_tests!(u128, test_u128);
    conversion_tests!(i128, test_i128);

    #[test]
    #[cfg(feature = "num-bigint")]
    fn test_bigint() {
        let i = BigInt::new(Sign::Plus, vec![5, 4, 8, 3, 2, 9, 3]);
        let s = SmallInt::from(i);
        assert_eq!(
            BigInt::from(s).to_u32_digits(),
            (Sign::Plus, vec![5, 4, 8, 3, 2, 9, 3])
        );
    }
}
